


# TODO: Move stuff related to "reading" in Tadpole quick start here 
# (e.g. file ops, result will be dataframe collection, dict or tuple)
# want data struct to share as input to build_anura (and abstract away where it came from, e.g. 
# xlsx file , csv files or other datatable etc)

